/**
   GRAB Graph Layout and Browser System

   Copyright (c) 1986, 1988 Regents of the University of California
   Copyright (c) 1989, Tera Computer Company
 **/

  /* io.c -- routines for reading and writing graph files */

#include <stdio.h>
#include <ctype.h>
#include "attribute.h"
#include "digraph.h"
#include "token.h"
#include "malloc.h" 

 /** 
    strings to signify separations between graph name, node declarations, and
    edge list
  **/
#define NAMESTART	"*NAME*"
#define NODESTART	"*NODES*"
#define EDGESTART	"*EDGES*"

#define NULL_FIELD "..."

  /**
     Format of files is as follows:  Fields are separated by whitespace
     (any combination of spaces and/or tabs).  Newlines indicate the end 
     of a node, edge, template, whatever and are not valid whitespace.
     Optional fields are indicated in brackets [].  If a field is included, 
     all previous fields on the line must also be included.  If no name 
     is given, the *NAME* line can be omitted.

        ...
        ...
     [*NAME*
     one line description of the graph]
     *NODES*
     distinguished_node_attribute_name [node_attribute_name ...]
     code [distinguished_node_attribute] [node_attribute ...] [shape] [brush] [color] [x y] [status] [level position]
        ...
        ...
     *EDGES*
     distinguished_edge_attribute_name [edge_attribute_name ...]
     from_code to_code [distinguished_edge_attribute] [edge_attribute ...] [brush] [color] [reversed]
        ...
        ...
   
     All lines before the *NAME* line are ignored.
     The distinguished attribute is used to label the node/edge.
     Extra attributes at the end of a line are ignored.  If no distinguished 
     attribute is given for a node, the code is used instead.  If 
     attributes are omitted for a particular node or edge they are considered 
     to be the null string.  If you wish to give a null field explicitly, 
     use the string "..."

     For valid shape, brush, color, and status strings, see the file
     attribute.c or the documentation.  reversed is a boolean value.
   **/

extern NODE *create_node();
extern BRUSH string_to_brush();
extern SHAPE string_to_shape();
extern COLOR string_to_color();
extern STATUS string_to_status();
extern BOOL string_to_bool();
extern char *color_to_string(), *brush_to_string(), *shape_to_string();
extern char *bool_to_string();
extern char *status_to_string();
extern TOKEN get_next_token(), token_val();

  /** 
     lint doesn't like the following too much, but it makes the code much
     more readable.
   **/

#define graph_token(f)  \
    (( (t = get_next_token(FALSE, FALSE)) == EOF_TOKEN) ? \
	eof_exit(f, "unexpected EOF encountered") : t)

BOOL read_nodes(), strip_nl();

static TOKEN t;
static FILE *iofile;

TOKEN eof_exit(fname, s)
char *fname, *s;
  /* unexpected end-of-file */
{
    fprintf(stderr, "%s: %s\n", fname, s);
    exit(1);
}

TOKEN bad_char_exit(s)
char *s;
  /* bad character read */
{
    print_bad_char(s);
    exit(1);
}

BOOL read_graph(digraph, infile)
DIGRAPH *digraph;
FILE *infile;
  /**
     read_graph reads in a list of nodes and edges from the ascii file with 
     file descriptor infile.  If the graph has been laid out,
     TRUE is returned; otherwise, FALSE is returned.
   **/
{
    BOOL is_laid_out;

    iofile = infile;
    init_reader(iofile);
    read_header(digraph);
    is_laid_out = read_nodes(digraph);
    read_edges(digraph);

      /**
         Remove edges going directly from node to node, when there
         is already a sequence of dummy edges connecting the nodes.
       **/
    remove_long_spans(digraph);
    return(is_laid_out);
} /* read_graph */

ignore_rest_of_line()
  /* just like it says */
{
    while ((t = graph_token("ignore_rest_of_line")) != NEWLINE_TOKEN)
    {
    }
}

read_header(digraph)
DIGRAPH *digraph;
  /** 
     read the information before the nodes
   **/
{
    TOKEN t;
    char name[MAXBUFFER];

      /* Skip comment lines, read in graph name if it exists */
    for (;;)
    {
	if ((t = graph_token("read_header")) == STRING_TOKEN && 
	     !strcmp(string_val, NAMESTART))
	{
	    ignore_rest_of_line();

	    if (fgets(name, MAXBUFFER, iofile) == NULL)
	      /** 
		 we want the whole line, white space and all, so we can't
		 use the token routines, which skip whitespace
	       **/
	    {
		PrintError("read_header: unexpected EOF encountered");
		exit(1);
	    }

	    if (!strip_nl(name))
	      /** 
		 strip_nl removes the newline which fgets puts at the end
		 of the string.  If there was no newline at the end, fgets
		 didn't finish reading the line, so the string wasn't long 
	         enough.  Read the rest of the line 
	       **/
	    {
		sprintf(stderr, "Graph name too long.  Only first %d characters retained.\n", MAXBUFFER-1);
		ignore_rest_of_line();
	    }
		
	    sprintf(Title(digraph), name);
	}
	else if (t == STRING_TOKEN && !strcmp(string_val, NODESTART))
	{
	    ignore_rest_of_line();
	    break;
	}
	else if (t != NEWLINE_TOKEN)
	{
	    ignore_rest_of_line();
	}
    }
}

TOKEN read_string(s, fname)
char *s, *fname;
  /**
     expect to read a string or a newline.
     If we get anything else, we aren't coming back.
     s should have room for MAXBUFFER characters, at least.
   **/
{
    if ((t = graph_token(fname)) == STRING_TOKEN || t == QSTRING_TOKEN)
    {
	strcpy(s, string_val);
	return t;
    }
    else if (t == NEWLINE_TOKEN)
    {
	strcpy(s, NULL_STRING);
	return t;
    }
    else if (t == ILLEGAL_CHAR_TOKEN)
    {
	char buf[MAXBUFFER];

	sprintf(buf, "%s: Illegal character:", fname);
	bad_char_exit(buf);
    }
    else
    {
	fprintf(stderr, "%s:  Unexpected token ", fname);
	print_token(t, TRUE);
	exit(1);
    }
}

BOOL read_nodes(digraph)
DIGRAPH *digraph;
  /**
     Read the node info and add it to digraph
     Return TRUE if the graph was laid out.  The graph is laid out if
     status and x y positions are supplied for all the nodes.
   **/
{
    char att_val[MAXBUFFER], code[MAXBUFFER];
    int x, y;
    char shapeName[MAXBUFFER], statusName[MAXBUFFER];
    char brushName[MAXBUFFER], colorName[MAXBUFFER];
    char xstr[MAXBUFFER], ystr[MAXBUFFER];
    char levelstr[MAXBUFFER], positionstr[MAXBUFFER];
    SHAPE shape;
    BRUSH brush;
    COLOR color;
    STATUS status;
    LNO level;
    int position;
    BOOL is_laid_out = TRUE;
    BOOL attr_nl;		/* was there a newline in the middle of 
					reading the attributes? */
    int num_attr;
    char **attr;
    int i;

    num_attr = read_node_template(digraph);

    for (;;)
      /* at start of loop:  we're at the start of the line */
    {
  	  /* set default values */
	attr_nl = FALSE;
	shape = DEFAULT_SHAPE;
	brush = DEFAULT_BRUSH;
	color = DEFAULT_COLOR;
	x = y = 0;
	status = DEFAULT_STATUS;
	level = -1;
	position  = -1;
	attr = (char **) malloc(num_attr * sizeof(char *));

	if (read_string(code, "read_nodes") == NEWLINE_TOKEN)
	  /* blank line */
	{
	    fprintf (stderr, 
	    	     "Unexpected blank line encountered while reading nodes\n");
	    continue;
	}
	else if (!strcmp(code, EDGESTART))
	  /* check for start of edge list */
	{
	    ignore_rest_of_line();
	    break;
	}

	if (read_string(att_val, "read_nodes") == NEWLINE_TOKEN)
	  /* nothing after the code */
	{
	      /* initialize the attributes */
	    for (i = 1; i < num_attr; i++)
	    {
		strsave(attr[i], NULL_STRING);
	    }

	      /* if the distinguished node attribute is omitted, use the code */
	    strsave(attr[0], code);
	    is_laid_out = FALSE;
	    (void) create_node(digraph, code, attr, shape, brush, color, TRUE,
	    	       	       x, y, status, level, position);
	    continue;
	}	
	else 	/* read in the attributes */
	{
	    if (!strcmp(att_val, NULL_FIELD))
	      /* next field should be null */
	    {
		strsave(attr[0], NULL_STRING);
	    }
	    else
	    {
		strsave(attr[0], att_val);
	    }
		
	    for (i = 1; i < num_attr; i++)
	    {
	        if (read_string(att_val, "read_nodes") == NEWLINE_TOKEN)
	          /* omitting some attributes */
		{
		    int j;

		    for (j = i; j < num_attr; j++)
		    {
		        strsave(attr[j], NULL_STRING);
		    }

		    attr_nl = TRUE;
		    break;
	        }
		else if (!strcmp(att_val, NULL_FIELD))
	          /* this field should be null */
	        {
		    strsave(attr[i], NULL_STRING);
	        }
	        else
	        {
		    strsave(attr[i], att_val);
	        }
	    }
	}

  	if (attr_nl || read_string(shapeName, "read_nodes") == NEWLINE_TOKEN)
	{
	    is_laid_out = FALSE;
	    (void) create_node(digraph, code, attr, shape, brush, color, TRUE,
	      		       x, y, status, level, position);
	    continue;
	}	

	shape = string_to_shape(shapeName);

	if (read_string(brushName, "read_nodes") == NEWLINE_TOKEN)
	{
	    is_laid_out = FALSE;
	    (void) create_node(digraph, code, attr, shape, brush, color, TRUE,
			       x, y, status, level, position);
	    continue;
	}	

	brush = string_to_brush(brushName);

	if (read_string(colorName, "read_nodes") == NEWLINE_TOKEN)
	{
	    is_laid_out = FALSE;
	    (void) create_node(digraph, code, attr, shape, brush, color, TRUE,
			       x, y, status, level, position);
	    continue;
	}	

	color = string_to_color(colorName);

	if (read_string(xstr, "read_nodes") == NEWLINE_TOKEN || 
	    read_string(ystr, "read_nodes") == NEWLINE_TOKEN)
	{
	    is_laid_out = FALSE;
	    (void) create_node(digraph, code, attr, shape, brush, color, TRUE,
			       x, y, status, level, position);
	    continue;
	}	

	x = atoi(xstr);
	y = atoi(ystr);

	if (read_string(statusName, "read_nodes") == NEWLINE_TOKEN)
	{
	    is_laid_out = FALSE;
	    (void) create_node(digraph, code, attr, shape, brush, color, TRUE,
			       x, y, status, level, position);
	    continue;
	}	

	status = string_to_status(statusName);

	if (read_string(levelstr, "read_nodes") == NEWLINE_TOKEN ||
	    read_string(positionstr, "read_nodes") == NEWLINE_TOKEN)
	{
	    (void) create_node(digraph, code, attr, shape, brush, color, TRUE,
			       x, y, status, level, position);
	    continue;
	}	

	level = atoi(levelstr);
	position = atoi(positionstr);

	(void) create_node(digraph, code, attr, shape, brush, color, TRUE,
		           x, y, status, level, position);

	ignore_rest_of_line();
    }

    return is_laid_out;
}

int read_node_template(digraph)
DIGRAPH *digraph;
  /* read the template for user-defined node attributes */
{
    int num_attr = 0;
    char **attr_names;
    char aname[MAXBUFFER];

    attr_names = (char **) malloc(sizeof(char *));

    while (read_string(aname, "read_node_template") == NEWLINE_TOKEN)
    {
	fprintf(stderr, "Unexpected blank line encountered while reading node attribute names\n");
    }

      /* read the distinguished attribute name */
    strsave(attr_names[num_attr], aname);
    num_attr++;

      /* read the rest of the attribute names */
    for (; ; num_attr++)
    {
	if (read_string(aname, "read_node_template") == NEWLINE_TOKEN)
	{
	    break;
	}	
	else
	{
            attr_names = (char **) realloc((char *) attr_names, 
					   (num_attr + 1) * sizeof(char *));
    	    strsave(attr_names[num_attr], aname);
	}
    }

    InitNodeAttrs(digraph, attr_names, num_attr, 0);
    return num_attr;
}

int read_edge_template(digraph)
DIGRAPH *digraph;
  /* read the template for user-defined edge attributes */
{
    int num_attr = 0;
    char **attr_names;
    char aname[MAXBUFFER];

    attr_names = (char **) malloc(sizeof(char *));

    while (read_string(aname, "read_edge_template") == NEWLINE_TOKEN)
    {
	fprintf(stderr, 
	    	"Unexpected blank line encountered while reading edge attribute names\n");
    }
    
      /* read the distinguished attribute name */
    strsave(attr_names[num_attr], aname);
    num_attr++;

      /* read the rest of the attribute names */
    for (; ; num_attr++)
    {
	if (read_string(aname, "read_edge_template") == NEWLINE_TOKEN)
	{
	    break;
	}	
	else
	{
            attr_names = (char **) realloc((char *) attr_names,
					   (num_attr + 1) * sizeof(char *));
    	    strsave(attr_names[num_attr], aname);
	}
    }

    InitEdgeAttrs(digraph, attr_names, num_attr, 0);
    return (num_attr);
}

read_edges(digraph)
DIGRAPH *digraph;
  /* read in the edge list. */
{
    TOKEN t;
    char att_val[MAXBUFFER];
    char from[MAXBUFFER], to[MAXBUFFER];
    char brushName[MAXBUFFER], colorName[MAXBUFFER];
    char reversedName[MAXBUFFER];
    BRUSH brush;
    COLOR color;
    int num_attr;
    char **attr;
    int i;
    BOOL attr_nl;		/* newline while reading attributes */
    BOOL reversed;

    num_attr = read_edge_template(digraph);

    for (;;)
      /* at start of loop:  we're at the start of the line */
    {
	  /* set default values */
	*from = *to = '\0';
	brush = DEFAULT_BRUSH;
	color = DEFAULT_COLOR;
	attr_nl = FALSE;
	reversed = FALSE;

	  /**
	     can't use read_string at the start of a line because an
	     eof is legal
	   **/
	if ((t = get_next_token(FALSE, FALSE)) == EOF_TOKEN)
	  /* end of file */
	{
	    break;
	}
	else if (t == STRING_TOKEN || t == QSTRING_TOKEN)
	{
	    strcpy(from, string_val);
	}
	else if (t == NEWLINE_TOKEN)
	{
	    fprintf(stderr, 
	    	    "Unexpected blank line encountered while reading edges\n");
	    continue;
	}
	else if (t == ILLEGAL_CHAR_TOKEN)
        {
	    bad_char_exit("read_edges:  Illegal character:");
        }
        else
        {
	    fprintf(stderr, "read_edges:  Unexpected token %d\n", t);
	    exit(1);
	}

	if (read_string(to, "read_edges") == NEWLINE_TOKEN)
	{
	    fprintf(stderr, "Missing to code for edge.  Ignoring this line\n");
	    ignore_rest_of_line();
	    continue;
	}

	attr = (char **) malloc(num_attr * sizeof(char *));

	  /**
	     read in the attributes.  The distinguished attribute is
	     optional for edges, so we don't need to do anything special
	     for it as we did for the nodes
	   **/
	for (i = 0; i < num_attr; i++)
	{
	    if (read_string(att_val, "read_edges") == NEWLINE_TOKEN)
	      /* omitting some attributes */
	    {
		int j;

		for (j = i; j < num_attr; j++)
		{
		    strsave(attr[j], NULL_STRING);
		}

		attr_nl = TRUE;
		break;
	    }
	    else if (!strcmp(att_val, NULL_FIELD))
	      /* this field should be null */
	    {
		strsave(attr[i], NULL_STRING);
	    }
	    else
	    {
		strsave(attr[i], att_val);
	    }
	}

   	if (attr_nl || read_string(brushName, "read_edges") == NEWLINE_TOKEN)
	{
	    (void) create_edge(digraph, from, to, attr, brush, color, reversed);
	    continue;
	}

	brush = string_to_brush(brushName);

	if (read_string(colorName, "read_edges") == NEWLINE_TOKEN)
	{
	    (void) create_edge(digraph, from, to, attr, brush, color, reversed);
	    continue;
	}

	color = string_to_color(colorName);
  
	if (read_string(reversedName, "read_edges") == NEWLINE_TOKEN)
	{
	    (void) create_edge(digraph, from, to, attr, brush, color, reversed);
	    continue;
	}

	reversed = string_to_bool(reversedName);

	(void) create_edge(digraph, from, to, attr, brush, color, reversed);
	ignore_rest_of_line();
    }
}

OUTEDGE *get_edge();

write_graph(digraph, outfile)
DIGRAPH *digraph;
FILE *outfile;
{
    NODE *node, *succ_node;
    VNO to_vno;
    OUTEDGE *edge;
    STATUS status, succ_status;
    int i;
    BOOL corh = FALSE;   	
      /**
	 have we seen a coalesced or hidden node? 
	 This is a hack, and something should be done about it.
	 The problem is that hidden and coalesced nodes aren't currently
	 marked as such when the graph is written out.  Instead, their
	 x_position and y_position are omitted, causing is_laid_out to
	 be false and the graph to be laid out when read in.  But the layout
	 routines don't work with dummy nodes, so they can't be written out
	 (nor can edges from or to them).

	 Note that after a hide or coalesce, the graph is always laid out
	 again, so no dummy nodes appear before hidden or coalesced nodes
	 (because the routines in relay.c remove all dummy nodes and
	 compact the nodes array.)  Therefore, if we set this variable when
	 we see a hidden or coalesced node, we can prevent the dummy nodes and
	 their edges from being written to the file
       **/

      /* print the graph title */
    if (Title(digraph)[0] != '\0')
    {
	fprintf(outfile, "%s\n", NAMESTART);
	fprintf(outfile, "%s\n", Title(digraph));
    }

      /* print node information */
    fprintf(outfile, "%s\n", NODESTART);

      /* print the node template */
    for (i = 0; i < NNodeAttr(digraph) - 1; i++)
    {
	fprintf(outfile, "\"%s\"\t", NAttrName(digraph, i));
    }

    fprintf(outfile, "%s\n", NAttrName(digraph, NNodeAttr(digraph) - 1));
	
    all_nodes(digraph, node)
    loop
	if (Coalescer(node))
	  /* don't write coalescer nodes */
	{
	   continue;
	}

	if (Is_dummy(node) && corh)
	  /* don't write dummy nodes if there are coalesced or hidden nodes */
	{
	    continue;
	}

	  /* print code (and tab) */
	fprintf(outfile, "\"%s\"\t", Name(node));

	  /* print attributes (with tabs) */
	for (i = 0; i < NNodeAttr(digraph); i++)
	{
	    if (!strcmp(NAttr(node, i), NULL_STRING))
	    {
	        fprintf(outfile, "%s\t", NULL_FIELD);
	    }
	    else
	    {
	        fprintf(outfile, "\"%s\"\t", NAttr(node, i));
	    }
	}

	  /* print shape string (and tab) */
	fprintf(outfile, "%s\t", shape_to_string(Shape(node)));

	  /* print brush string (and tab) */
	fprintf(outfile, "%s\t", brush_to_string(Brush(node)));

	  /* print color string (and tab) */
	fprintf(outfile, "%s\t", color_to_string(Color(node)));

	if (Displayed(node) && !Coalesced(node))
	{
	      /* print x and y positions (with tabs) */
	    fprintf(outfile, "%d\t%d\t", X_position(node), Y_position(node));

	      /* print what type of node it is */
	    fprintf(outfile, "%s\t", status_to_string(Status(node)));

	      /* print the level number and position, with a \n at end */
	    fprintf(outfile, "%d\t%d\n", Level_no(node), Position(node));
	}
	else
	  /* coalesced or hidden node; don't print out layout info */
	{
	    corh = TRUE;

	    fprintf(outfile, "\n");
	}
    endloop

      /* print edge information */
    fprintf(outfile, "%s\n", EDGESTART);

      /* print edge attribute template */
    for (i = 0; i < NEdgeAttr(digraph) - 1; i++)
    {
	fprintf(outfile, "\"%s\"\t", EAttrName(digraph, i));
    }

    fprintf(outfile, "%s\n", EAttrName(digraph, NEdgeAttr(digraph) - 1));
	
    all_nodes(digraph, node)
    loop
	  /* Print all edges in out edge list */
	status = Status(node);

	all_out_edges(node, edge)
	loop
	    succ_node = Node(digraph, To_vno(edge));

	    fprintf(outfile, "\"%s\"\t", Name(node));
	    fprintf(outfile, "\"%s\"\t", Name(succ_node));

	      /* print attributes (with tabs) */
	    for (i = 0; i < NEdgeAttr(digraph); i++)
	    {
	        if (!strcmp(EAttr(edge, i), NULL_STRING))
	        {
	            fprintf(outfile, "%s\t", NULL_FIELD);
	        }
	        else
	        {
	            fprintf(outfile, "\"%s\"\t", EAttr(edge, i));
	        }
	    }

	      /* print brush (and tab) */
	    fprintf(outfile, "%s\t", brush_to_string(Brush(edge)));

	      /* print color (and tab) */
	    fprintf(outfile, "%s\t", color_to_string(Color(edge)));

	      /* print reversed (and newline) */
	    fprintf(outfile, "%s\n", bool_to_string(Edge_reversed(edge)));
	endloop

        if (!corh)
	  /**
	     Print remaining edges which go to or from 
	     dummies, using information in the successor sets.
	     Note that no attributes are available for these edges.
	   **/
	{
	    each_element(Succ_set(node), to_vno)
	    loop
	        OUTEDGE *edge;
	        int j;
    
	        succ_node = Node(digraph, to_vno);
	        succ_status = Status(succ_node);
       
	        if (Coalescer(succ_node) || 
		    (status != DUMMY && succ_status != DUMMY))
		  /* don't print edges from coalescer or normal nodes */
	        {
	            continue;
	        }
        
	        for (j = max_edges(node, succ_node); j > 0; j--)
	        {
	            if ((edge = get_edge(digraph, node, succ_node, j)) != NULL)
		    {
	                fprintf(outfile, "%s\t", Name(node));
	                fprintf(outfile, "%s\t", Name(succ_node));
            
	                  /* print attributes (and tabs) */
	                  /** For now, print nothing (no values assigned) **/
	                for (i = 0; i < NEdgeAttr(digraph); i++)
	                {
	                    fprintf(outfile, "%s\t", NULL_FIELD);
	                }
        
	                  /* print brush (and tab) */
	                fprintf(outfile, "%s\t", brush_to_string(Brush(edge)));
            
	                  /* print color (and tab) */
	                fprintf(outfile, "%s\t", color_to_string(Color(edge)));
            
	                  /* print reversed (and newline) */
	                fprintf(outfile, "%s\n", 
			        bool_to_string(Edge_reversed(edge)));
		    }
	        }
	    endloop
	}
    endloop
} /* write_graph */

BOOL strip_nl(s)
char *s;
  /* strip the newline, if any, off the end of a string. */
{
    int i;

    i = strlen(s) - 1;

    if (s[i] == '\n')
    {
	s[i] = '\0';
	return TRUE;
    }
    else
    {
	return FALSE;
    }
}
