/**
   GRAB Graph Layout and Browser System

   Copyright (c) 1986, 1988 Regents of the University of California
   Copyright (c) 1989, Tera Computer Company
 **/

  /**
     mincross.c contains procedures to minimize the number of edge crossings
     in a digraph. 
   **/

#include "attribute.h"
#include "digraph.h"
#include "debug.h"
#include "heuristic.h"

BOOL sort_by_bc();
BOOL sort_level();
BOOL reverse_equal_bc();
BOOL reverse_level();
float get_bc();

minimize_crossings(digraph)
DIGRAPH *digraph;
  /**
     minimize_crossings uses the barycentric method to minimize the number
     of edge crossings in digraph. 
   **/
{
    int nc;			/* final number of crossings found */

    compute_all_positions(digraph);	/* compute the initial positions */
    compute_barycenters(digraph);	/* compute the initial barycenters */

    if (debug)
    {
	printf("minimize_crossing:  Initial graph\n");
	offset_levels(digraph);
    }

    nc = reduce_crossings(digraph);	 /* minimize crossings */

    if (debug2)
    {
	printf("Final number of arc crossings: %d\n", nc);
    }
}				

reduce_crossings(digraph)
DIGRAPH *digraph;
  /**
     reduce_crossings uses the barycentric method to reduce the number of
     edge crossings in digraph. 
   **/
{
    int nc_up, nc_down, nc_up_down;	/* number of crossings after sort */

    if (debug2)
    {
	printf("Sorting by UP barycenters...\n");
    }

    (void) sort_by_bc(digraph, UP);	/* sort each level by up-barycenters */

    if (debug2)
    {
	printf("Sorting by DOWN barycenters...\n");
    }

    (void) sort_by_bc(digraph, DOWN);  /* sort each level by down-barycenters */

    if (debug2)
    {
	printf("Sorting by UP barycenters...\n");
    }

    (void) sort_by_bc(digraph, UP);	/* sort each level by up-barycenters */
    nc_up = num_crossings(digraph);

    if (debug2)
    {
	printf("\tCrossings so far: %d\n", nc_up);
    }

    if (debug2)
    {
	printf("Sorting by DOWN barycenters...\n");
    }

    (void) sort_by_bc(digraph, DOWN);  /* sort each level by down-barycenters */
    nc_down = num_crossings(digraph);

    if (debug2)
    {
	printf("\tCrossings so far: %d\n", nc_down);
    }

    if (nc_down == 0)
    {
	return (nc_down);
    }

    if (nc_down > nc_up)
    {
	if (debug2)
	{
	    printf("Sorting by UP barycenters...\n");
 	}

	(void) sort_by_bc(digraph, UP);
	nc_up = num_crossings(digraph);

	if (debug2)
	{
	    printf("\tCrossings so far: %d\n", nc_up);
	}

	if (nc_up == 0)
	{
	    return (nc_up);
	}
    }

    if (debug2)
    {
	printf("Sorting by UP/DOWN barycenters...\n");
    }

    (void) sort_by_bc(digraph, UPDOWN);
    nc_up_down = num_crossings(digraph);

    if (debug2)
    {
	printf("\tCrossings so far: %d\n", nc_up_down);
    }

    if (nc_up_down == 0)
    {
	return (nc_up_down);
    }

    if (nc_up_down > nc_up && nc_down >= nc_up)
    {
	if (debug2)
	{
	    printf("Sorting by DOWN barycenters...\n");
	}

	(void) sort_by_bc(digraph, DOWN);

	if (debug2)
	{
	    printf("Sorting by UP barycenters...\n");
	}

	(void) sort_by_bc(digraph, UP);
	nc_up = num_crossings(digraph);

	if (debug2)
	{
	    printf("\tCrossings so far: %d\n", nc_up);
	}

	return (nc_up);
    }

    if (nc_up_down > nc_down && nc_up >= nc_down)
    {
	if (debug2)
	{
	    printf("Sorting by UP barycenters...\n");
	}

	(void) sort_by_bc(digraph, UP);

	if (debug2)
	{
	    printf("Sorting by DOWN barycenters...\n");
	}

	(void) sort_by_bc(digraph, DOWN);
	nc_down = num_crossings(digraph);

	if (debug2)
	{
	    printf("\tCrossings so far: %d\n", nc_down);
	}

	return (nc_down);
    }

    return (nc_up_down);
}				/* minimize_crossings */

compute_all_positions(digraph)
DIGRAPH *digraph;
  /**
     compute_all_positions computes the positions of the members of each
     level of digraph. 
   **/
{
    LEVEL *level;		/* each level */

    each_level(digraph, level)
    loop
	compute_positions(level);
    endloop
}				/* compute_all_positions */

compute_barycenters(digraph)
DIGRAPH *digraph;
  /**
     compute_barycenters computes the up and down barycenters for each node
     of each level of digraph. 
   **/
{
    LEVEL *level;		/* each level */
    MEMBER *member;		/* each member */

    each_level(digraph, level)
    loop
	each_member(level, member)
	loop
	    compute_bc(digraph, member, UP);
     	    compute_bc(digraph, member, DOWN);
        endloop
    endloop
}				/* compute_barycenters */

compute_positions(level)
LEVEL *level;			/* level of nodes */
  /* compute_positions computes the position of each member of level. */
{
    int position;		/* each position */
    MEMBER *member;		/* each member */

    if (debug)
    {
	(void) fix_positions(level);
	return;
    }

    position = 1;		/* first node is at position 1 */

    each_member(level, member)
    loop
	X_position(member) = position++;
    endloop
}				/* compute_positions */

compute_bc_level(digraph, level, direction)
DIGRAPH *digraph;
LEVEL *level;
DIRECTION direction;
  /**
     compute_bc_level computes the bc's for each member of level for
     direction direction. 
   **/
{
    MEMBER *member;		/* each member */

    each_member(level, member)
    loop
	compute_bc(digraph, member, direction);
    endloop
}				/* compute_bc_level */

extern NODE *get_next_node(), *get_prev_node();

compute_bc(digraph, member, direction)
DIGRAPH *digraph;
MEMBER *member;
DIRECTION direction;		/* UP or DOWN */
  /**
     compute_bc computes the up or down barycenter of member, according to
     the value of direction.  If member has no such barycenter, then
     Bc(member, direction) is set to NO_BC. 
   **/
{
    SET *vno_set;	/* set of antecedents or succedents */
    VNO vno;		/* each antecedent or succedent */
    int num;		/* number of antecedent or succedent nodes */
    int weight;		/* total weight of antecedent or succedent nodes */
    NODE *immediate, *realnode;

    switch (direction)
    {
        case UP:			/* compute up_bc */
	    vno_set = Ante_set(Member_node(member));
	    break;
	case DOWN:			/* compute down_bc */
	    vno_set = Succ_set(Member_node(member));
	    break;
    }

    num = 0;			/* initialize */
    weight = 0;

    each_element(vno_set, vno)
    loop
	num++;			/* another antecedent/succedent */
	  /**
	     Code added 2/21/86 by ebm.  Compute the barycenter as some
	     percentage of the average positions of the immediate pred/succ's
	     plus a percentage of the postions of the Actual nodes that are the
	     pred/succ's. 
	   **/
	immediate = Node(digraph, vno);
	weight += immedPercent * X_position(Node_member(immediate));

	if (Is_dummy(immediate))
	{
	    switch (direction)
	    {
		case UP:
		    realnode = Node_member(get_prev_node(digraph, immediate));
		    break;
		case DOWN:
		    realnode = Node_member(get_next_node(digraph, immediate));
		    break;
	    }
	}
	else
	{
	    realnode = immediate;
	}

	weight += (1.0 - immedPercent) * X_position(Node_member(realnode));
    endloop

    if (num == 0)		/* no barycenter */
    {
	Bc(member, direction) = NO_BC;
    }
    else			/* compute the bc */
    {
	Bc(member, direction) = (BC) weight / (BC) num;
    }
} /* compute_bc */

BOOL sort_by_bc(digraph, direction)
DIGRAPH *digraph;
DIRECTION direction;
  /**
     sort_by_bc sorts the nodes in each level of digraph in order of non-
     decreasing up (direction == UP) or down (direction == down)
     barycenters. If any of the levels were out of order, TRUE is returned;
     FALSE otherwise. The bc's of the previous and next levels of each level
     are recomputed if the node order of the level is changed. 
   **/
{
    LEVEL *level;		/* each level */
    BOOL change;		/* TRUE if the order is changed on any
				 * level */

    change = FALSE;		/* assume all levels are sorted */

    switch (direction)
    {
        case UP:		/* sort from first level to last */
	    each_level(digraph, level)
	    loop
	        if (Prev_level(level) != NULL)	/* there is an up bc */
		{
	            change |= sort_level(digraph, level, direction);
		}
	    endloop
	    break;
        case DOWN:		/* sort from last level to first */
	    each_level_reverse(digraph, level)
	    loop
	        if (Next_level(level) != NULL)	/* there is a down bc */
		{
	            change |= sort_level(digraph, level, direction);
		}
	    endloop
	    break;
        case UPDOWN:		/* sort both ways by average of up and down */
	    each_level(digraph, level)
	    loop
	        change |= sort_level(digraph, level, direction);
	    endloop
	    break;
    }

    if (debug)
    {
	printf("\nsort_by_bc:  %s, %d crossings\n",
	       (direction == UP) ? "UP" : ((direction == DOWN) ?
					   "DOWN" : "UPDOWN"), 
	       num_crossings(digraph));
    }

    return (change);
}				/* sort_by_bc */

BOOL sort_level(digraph, level, direction)
DIGRAPH *digraph;
LEVEL *level;
DIRECTION direction;
  /**
     sort_level sorts the nodes in level in order of non- decreasing up
     (direction == UP) or down (direction == down) barycenters. If level was
     out of order, TRUE is returned; FALSE otherwise. The bc's of the
     previous and next levels of level are recomputed if the node order of
     the level is changed. 
   **/
{
    MEMBER *member;		/* each member */
    MEMBER *prev;		/* each previous member */
    BOOL change;		/* TRUE if the order is changed on any level */
    float position_ratio;	/* ratio of number of members in levels */

    change = FALSE;		/* assume nodes are sorted */

    switch (direction)
    {
        case UP:
	    if (Prev_level(level) == NULL)
	    {
	        position_ratio = 1.00;
	    }
	    else
	    {
	        position_ratio = ((float) card(Members(Prev_level(level)))) /
		                 ((float) card(Members(level)));
	        break;
	    }
        case DOWN:
	    if (Next_level(level) == NULL)
	    {
	        position_ratio = 1.00;
	    }
	    else
	    {
	        position_ratio = ((float) card(Members(Next_level(level)))) /
		                 ((float) card(Members(level)));
	        break;
	    }
        case UPDOWN:
	    if (Prev_level(level) == NULL || Next_level(level) == NULL)
	    {
	        position_ratio = 1.00;
	    }
	    else
	    {
	        position_ratio = ((float) card(Members(Prev_level(level)))) /
		                 ((float) card(Members(Next_level(level))));
	        break;
	    }
    }

    each_member(level, member)
    loop
	for (prev = Prev_member(member); 
	     (prev != NULL) && (get_bc(prev, direction, position_ratio) >
	      			get_bc(member, direction, position_ratio));
	     prev = Prev_member(prev))
	{}

        if (prev != Prev_member(member))	/* move to new position */
        {
	    change = TRUE;

	      /* first unlink from previous member */
	    Next_member(Prev_member(member)) = Next_member(member);

	    if (Next_member(member) != NULL)	/* unlink from next */
	    {
	        Prev_member(Next_member(member)) = Prev_member(member);
	    }
    
	    if (prev == NULL)	/* insert at the head */
	    {
	        Prev_member(Order(level)) = member;
	        Next_member(member) = Order(level);
	        Prev_member(member) = NULL;
	        Order(level) = member;
	    }
	    else			/* insert after prev */
	    {
	        Next_member(member) = Next_member(prev);

	        if (Next_member(member) != NULL)
		{
		    Prev_member(Next_member(member)) = member;
		}

	        Prev_member(member) = prev;
	        Next_member(prev) = member;
	    }
        }
    endloop

    if (change)		
      /* recompute positions on level and bc's of affected levels */
    {
	compute_positions(level);
	Num_cross(level) = -1;	/* num crossings no longer valid */

	if (Prev_level(level) != NULL)	/* recompute down-bc of prev level */
	{
	    compute_bc_level(digraph, Prev_level(level), DOWN);
	    Num_cross(Prev_level(level)) = -1;	
	      /* num crossings no longer valid */
	}

	if (Next_level(level) != NULL)	/* recompute up-bc of next level */
	{
	    compute_bc_level(digraph, Next_level(level), UP);
	}
    }

    if (debug)
    {
	printf("\nsort_level:  %s, %d crossings\n",
	       (direction == UP) ? "UP" : "DOWN", num_crossings(digraph));
    }

    return (change);
} /* sort_level */

BOOL reverse_equal_bc(digraph, direction)
DIGRAPH *digraph;
DIRECTION direction;
  /**
     reverse_equal_bc reverses the nodes with equal barycenters in the first
     level of digraph which has nodes with equal barycenters. 
   **/
{
    LEVEL *level;		/* each level */
    BOOL change;		/* TRUE if the order is changed on any level */

    change = FALSE;		/* assume all levels will be changed */

    switch (direction)
    {
        case UP:		/* reverse from first level to last */
   	    each_level(digraph, level)
	    loop
	        if (Prev_level(level) != NULL)	/* there is a up bc */
	  	{
	            change |= reverse_level(digraph, level, direction);

	      	    if (change)		        /* don't continue */
		    {
		        break;
		    }
	        }
	    endloop
	    break;
        case DOWN:		/* reverse from last level to first */
	    each_level_reverse(digraph, level)
	    loop
	        if (Next_level(level) != NULL)	/* there is a down bc level */
	        {
	            change |= reverse_level(digraph, level, direction);

	            if (change)			/* don't continue */
		    {
		        break;
		    }
	        }
	    endloop
	    break;
    }

    if (debug)
    {
 	printf("\nreverse_equal_bc:  %s, %d crossings\n",
               (direction == UP) ? "UP" : "DOWN", num_crossings(digraph));
    }

    return (change);
} /* reverse_equal_bc */

BOOL reverse_level(digraph, level, direction)
DIGRAPH *digraph;
LEVEL *level;
DIRECTION direction;
  /**
     reverse_level reverses the nodes in level with equal barycenters, if
     any. If any nodes were reversed, TRUE is returned; FALSE otherwise. The
     bc's of the previous and next levels of level are recomputed if the
     node order of the level is changed. 
   **/
{
    MEMBER *start, *end;	/* start & end of an equal bc list */
    MEMBER *prev_member;	/* previous member to start */
    MEMBER *next_member;	/* next member to end */
    BOOL change;		/* TRUE if the order is changed on level */

    change = FALSE;		/* assume no equal barycenters */

    each_member(level, start)
    loop
	if ((Bc(start, direction) == NO_BC) || (Next_member(start) == NULL))
	{
	    continue;
	}

        if (Bc(start, direction) == Bc(Next_member(start), direction))
	{
	    break;
	}
    endloop

    for (end = start; 
	 (end != NULL) && (Next_member(end) != NULL) &&
	   (Bc(end, direction) == Bc(Next_member(end), direction));
	 end = Next_member(end))
    {}

    if (start != end)		/* there is a run of equal barycenters */
    {
	change = TRUE;
	prev_member = Prev_member(start);
	next_member = Next_member(end);

	reverse_list(start, end);	/* reverse the list */

 	  /* now start points to the end of the list, and end to the start */
	Prev_member(end) = prev_member;
	if (prev_member != NULL)
	{
	    Next_member(prev_member) = end;
	}
	else			/* new head */
	{
	    Order(level) = end;
	}

	Next_member(start) = next_member;

	if (next_member != NULL)
	{
	    Prev_member(next_member) = start;
	}

	compute_positions(level);	/* recompute positions */

	  /* now recompute bc's of affected levels */
	if (Prev_level(level) != NULL)	/* recompute down-bc of prev level */
	{
	    compute_bc_level(digraph, Prev_level(level), DOWN);
	}

	if (Next_level(level) != NULL)	/* recompute up-bc of next level */
	{
	    compute_bc_level(digraph, Next_level(level), UP);
	}
    }

    return (change);
} /* reverse_level */

reverse_list(start, end)
MEMBER *start, *end;
  /* reverse_list reverses the members between start and end. */
{
    MEMBER *head;		/* head of the list */

    if (start == end)
    {
	return;
    }

    head = Next_member(start);
    reverse_list(head, end);
    Next_member(head) = start;
    Prev_member(start) = head;
}				/* reverse_list */

num_crossings(digraph)
DIGRAPH *digraph;
  /* num_crossings computes the number of edge crossings in digraph. */
{
    LEVEL *level;		/* each level */
    MEMBER *member;		/* each member */
    MEMBER *next_member;	/* each member after member */
    VNO succ_vno;		/* each succedent to member */
    int succ_position;		/* position of succ_vno */
    VNO vno;			/* each succedent to next_member */
    int count;			/* number of crossings */

    count = 0;			/* initialize */

    each_level(digraph, level)
    loop
	if (Next_level(level) == NULL)
	{
	    return (count);
	}

        if (Num_cross(level) != -1)
        {
	    count += Num_cross(level);
	    continue;
        }
        Num_cross(level) = 0;

        each_member(level, member)
	loop
	    each_element(Succ_set(Member_node(member)), succ_vno)
	    loop
	        succ_position = X_position(Member(digraph, succ_vno));

                if (succ_position == 1)	/* no crossings from this node */
		{
		    continue;
		}

    		for (next_member = Next_member(member); next_member != NULL;
	 	     next_member = Next_member(next_member))
    		{
		    each_element(Succ_set(Member_node(next_member)), vno)
	    	    loop
	    		if (X_position(Member(digraph, vno)) < succ_position)
			{
	    		    count++;
	    		    (Num_cross(level))++;
			}
		    endloop
    		}
    	    endloop
	endloop
    endloop

    return (count);
}				/* num_crossings */

float get_bc(member, direction, position_ratio)
MEMBER *member;
DIRECTION direction;
float position_ratio;
  /**
     get_bc returns the up or down bc of a node, if it has one.  If a node
     doesn't have a bc, get_bc estimates one based on the current position
     and on the ratio of the number of members in the current level to the
     number of members in the previous or next level, as appropriate. 
   **/
{
    float bc, upbc, downbc;	/* barycenter */
    int denom;

    switch (direction)
    {
        case UP:
        case DOWN:
	    bc = Bc(member, direction);

	    if (bc != NO_BC)	/* has barycenter -- return actual bc */
	    {
	        return (bc);	
	    }
	    else		/* no barycenter -- return estimated bc */
	    {
	        return ((float) X_position(member) * position_ratio);
	    }

        case UPDOWN:
	    denom = card(Ante_set(member)) + card(Succ_set(member));

	    if (denom == 0)
	    {
	        bc = 0;
	    }
	    else
	      /**
	         Code added 2/20/86 by ebm.  Mdavis original code (below)
	         calculated the UPDOWN bc using the position ratio.  By
	         invoking the option -h0, this can be avoided.  Also, the
	         original code didn't check if the component UP/DOWN bc's
	         existed. 
	       **/
	    {
	        if ((upbc = Bc(member, UP)) == NO_BC)
		{
		    upbc = (float) X_position(member) * position_ratio;
		}

	        if ((downbc = Bc(member, DOWN)) == NO_BC)
		{
		    downbc = (float) X_position(member) * position_ratio;
		}

	        if (heuristic[0])
		{
		    bc = (card(Ante_set(member)) * upbc
		           + card(Succ_set(member)) * downbc)
		         / denom;
		}
	        else
		{
		      /** mdavis original code **/
		    bc = (card(Ante_set(member)) * upbc + 
			  card(Succ_set(member)) * downbc * position_ratio) / 
			 denom;
		}
	    }

	    return (bc);

	default:
	    return (0.0);
    }
} /* get_bc */
